/*
  Copyright (c) 2008 Seneca College
  Licenced under the MIT License (http://www.c3dl.org/index.php/mit-license/)
  Author: Patrick Lam
*/

/**
	@class A Picking class
*/
function Picking(scene)
{
	var debug = "";
	var debugField;
	var scn = new Scene();
	scn = scene.getScene();
	var cam = new FreeCamera();
	cam = scn.getCamera();
	var TEST_CULL = false;	// culling in triangle intersection test (vertex culling?)

	this.setScene = function(scene)
	{
		scn = scene;
	}
	
	// ====================================================
	// For debugging
	function displayDebug(message) {
		debugField.value = message;
	}

	this.setTextField = function(field) {
		debugField = field;
	}
	// ====================================================

	this.onMouseUp = function(event)
	{
	}

	/**
		Returns an array of object index number in the current scene in sorted order (closet to furthest)

		@returns {Array} The array of object index number.
	*/
	this.onMouseDown = function(event)
	{
		debug = "";
		// Button    Which    Button
		// Left  	 1		  0
		// Middle 	 2		  1
		// Right 	 3  	  2

		//==========================================================================
		// Disables Right Click Menu - copied from internet
		var message="Right Clicked";

		function clickIE(){
			if (document.all){
				(message);
				return false;
			}
		}

		function clickNS(e) {
			if(document.layers||(document.getElementById&&!document.all)){
				if (e.which==2||e.which==3) {
					alert(message);
					return false;
				}
			}
		}

		if (document.layers){
			document.captureEvents(Event.MOUSEDOWN);
			document.onmousedown = clickNS;
		}
		else{
			document.onmouseup = clickNS;
			document.oncontextmenu = clickIE;
		}

		document.oncontextmenu = new Function("return false")
		//==========================================================================

		// getElementsByTagName returns an array
		// Take the first canvas tag
		var canvasTag = document.getElementsByTagName('canvas')[0];

		// Mouse Coord on canvas
		var canvasXY = getXandY(event);

		// Camera Position
		var camPos = cam.getPosition();

		// Find dimension of far clipping plane
		var farWidth = 2 * ((C3DL_FAR_CLIPPING_PLANE) * Math.tan(degreesToRadians(0.5 * C3DL_FIELD_OF_VIEW)));
		var farHeight = farWidth * (canvasTag.height/canvasTag.width);

		// Find ratio between far clipping plane and canvas screen
		var wRatio = farWidth / canvasTag.width;		
		var hRatio = farHeight / canvasTag.width;

		// Find point clicked on far clipping plane
		var farX = (canvasXY[0] / canvasTag.width) * farWidth;
		var farY = (canvasXY[1] / canvasTag.height) * farHeight;
		var camDist = Math.sqrt(camPos[0] * camPos[0] + camPos[1] * camPos[1] + camPos[2] * camPos[2]);
		// Since we are starting off with positions aligned on the z-axis
		// The camera is hypothetically on the z-axis, and so, the z value should be
		// the distance of the camera from the origin of the axis and not just the
		// the z value (projection on z-axis) of the camera's position
		var farZ = C3DL_FAR_CLIPPING_PLANE - camDist;

		// Create the Starting point and direction of mouse vector
		var mouseOrigin = camPos;
		var mouseVec = makeVector(farX, farY, farZ);

		// View Matrix
		viewMatrix = makePoseMatrix(cam.getLeft(), cam.getUp(), cam.getDir(), cam.getPosition());

		// Move mouse vector to align with camera
		mouseVec = multiplyMatrixByVector(viewMatrix, mouseVec);

		// Array holding index of objects that pass the bounding box test
		var passRayBoxTest = new Array();

		// Find all objects with bounding box intersecting with mouse vector
		for (var i = 0; i < scn.getObjListSize(); i++) {
			var testObj = scn.getObj(i);
			var boxValues = testObj.getBoundingBox(true);
			// getBoundingBox() returns {minX, minY, minZ, maxX, maxY, maxZ} that has not been transformed
			// only gets scaled
			var minVal = new Array(boxValues[0], boxValues[1], boxValues[2]);
			var maxVal = new Array(boxValues[3], boxValues[4], boxValues[5]);

			// Move the mouse vector into the object's space
			var objTransMat = makePoseMatrix(testObj.getLeft(), testObj.getUp(), testObj.getDirection(), testObj.getPosition());
			var objInvTransMat = inverseMatrix(objTransMat);
			var tempOrigin = multiplyMatrixByVector(objInvTransMat, mouseOrigin);
			var tempVec = multiplyMatrixByVector(objInvTransMat, mouseVec);
				
			if (CheckLineBox(minVal, maxVal, tempOrigin, tempVec)) {
				passRayBoxTest.push(i);
			}
		}

		// Temporary until ray-triangle test is complete
		var intersectObjs = passRayBoxTest;
		/*
		// ===========================================================================
		//	FOR RAY-TRIANGLE INTERSECTION TEST, STILL WORKING ON IT
		// ===========================================================================
		// Array holding index of objects that pass the triangle test
		var intersectObjs = new Array();

		// Find all objects with triangles(faces) intersecting with mouse vector
		for (var i = 0; i < passRayBoxTest.length; i++) {
			var testObj = scn.getObj(passRayBoxTest[i]);			
			var vertices = testObj.getVertices();
			var passTest = false;
			
			// Objects scale
			var wtScale = new Array( testObj.getScale()[0],0,0,0,
									 0,testObj.getScale()[1],0,0,
									 0,0,testObj.getScale()[2],0,
									 0,0,0,1);	

			// Go through the object's faces
			for (var j = 0; j < vertices.length && !passTest; j += 9) {
				// 3 points of a triangle
				var vert1 = new Array(vertices[j], vertices[j+1], vertices[j+2]);
				var vert2 = new Array(vertices[j+3], vertices[j+4], vertices[j+5]);
				var vert3 = new Array(vertices[j+6], vertices[j+7], vertices[j+8]);

				// Scale the vertices
				vert1 = multiplyMatrixByVector(wtScale, vert1);
				vert2 = multiplyMatrixByVector(wtScale, vert2);
				vert3 = multiplyMatrixByVector(wtScale, vert3);

				// Move the mouse vector into object's space
				var objTransMat = makePoseMatrix(testObj.getLeft(), testObj.getUp(), testObj.getDirection(), testObj.getPosition());
				var objInvTransMat = inverseMatrix(objTransMat);
				var tempOrigin = multiplyMatrixByVector(objInvTransMat, mouseOrigin);
				var tempVec = multiplyMatrixByVector(objInvTransMat, mouseVec);

				if (intersect_triangle2(tempOrigin, tempVec, vert1, vert2, vert3) == true) {
					intersectObjs.push(passRayBoxTest[i]);
					passTest = true;
				}
			}
		}
		// ==================================================================================
		*/

		// Sort all intersecting objects from closet to farthest 
		for (var i = 0; i < intersectObjs.length; i++) {
			for (var j = 0; j < intersectObjs.length; j++) {
				// Objects Positions
				var objAPos = scn.getObj(intersectObjs[i]).getPosition();
				var objBPos = scn.getObj(intersectObjs[j]).getPosition();

				// Objects x,y,z distances
				var dxA = camPos[0] - objAPos[0];
				var dyA = camPos[1] - objAPos[1];
				var dzA = camPos[2] - objAPos[2];

				var dxB = camPos[0] - objBPos[0];
				var dyB = camPos[1] - objBPos[1];
				var dzB = camPos[2] - objBPos[2];

				// Objects distance from camera
				var distA = Math.sqrt(dxA * dxA + dyA * dyA + dzA * dzA);
				var distB = Math.sqrt(dxB * dxB + dyB * dyB + dzB * dzB);

				// Swap
				if (distA < distB) {
					var tmp = intersectObjs[i];
					intersectObjs[i] = intersectObjs[j];
					intersectObjs[j] = tmp;	
				}
			}
		}
		return intersectObjs;
	}

	// -------------------------------------------------------------------------
	// Mouse Position Functions

	// Find position of object in browser
	// Returning it's offset in the browser
	function findPos(obj) {
		var curleft = curtop = 0;
		if (obj.offsetParent) {
			do {
				curleft += obj.offsetLeft;
				curtop += obj.offsetTop;
			} while (obj = obj.offsetParent);
			return [curleft,curtop];
		}
	}

	// Returns the mouse x,y in canvas
	function getXandY( event )
	{
		var tempDiv = document.getElementsByTagName('canvas')[0];
		var coords = findPos(tempDiv);

		// determine the correct X, Y
		var X = ( ( event.clientX - coords[0] ) + window.pageXOffset )  - (tempDiv.width / 2.0);
		var Y = (tempDiv.height / 2.0) - ( event.clientY - ( coords[1] - window.pageYOffset ) );

		return [X,Y];
	}

	// -------------------------------------------------------------------------
	// Ray-Bounding Box Intersection Test
	// Modified Version of Sample function from 3D Programming
	function GetIntersection(fDst1, fDst2, P1, P2) {
		if ((fDst1 * fDst2) >= 0) return 0;
		if ( fDst1 == fDst2 ) return 0;

		var Hit = new Array();

		Hit[0] = P1[0] + (P2[0]-P1[0]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );
		Hit[1] = P1[1] + (P2[1]-P1[1]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );
		Hit[2] = P1[2] + (P2[2]-P1[2]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );

		return Hit;
	}

	function InBox( Hit, B1, B2, Axis) {
		if (Hit != 0) {
			if ( Axis==1 && Hit[2] > B1[2] && Hit[2] < B2[2] && Hit[1] > B1[1] && Hit[1] < B2[1]) return 1;
			if ( Axis==2 && Hit[2] > B1[2] && Hit[2] < B2[2] && Hit[0] > B1[0] && Hit[0] < B2[0]) return 1;
			if ( Axis==3 && Hit[0] > B1[0] && Hit[0] < B2[0] && Hit[1] > B1[1] && Hit[1] < B2[1]) return 1;
		}	

		return 0;
	}

	// B1 is array of 3 values, the min X,Y,Z values of bounding box
	// B2 is array of 3 values, the max X,Y,Z values of bounding box
	// L1 is array of 3 values, the X,Y,Z values of starting position of ray
	// L2 is array of 3 values, the X,Y,Z values of direction vector of ray	
	function CheckLineBox(B1, B2, L1, L2) {
		if (L2[0] < B1[0] && L1[0] < B1[0]) return false;
		if (L2[0] > B2[0] && L1[0] > B2[0]) return false;
		if (L2[1] < B1[1] && L1[1] < B1[1]) return false;
		if (L2[1] > B2[1] && L1[1] > B2[1]) return false;
		if (L2[2] < B1[2] && L1[2] < B1[2]) return false;
		if (L2[2] > B2[2] && L1[2] > B2[2]) return false;
		if (L1[0] > B1[0] && L1[0] < B2[0] &&
			L1[1] > B1[1] && L1[1] < B2[1] &&
			L1[2] > B1[2] && L1[2] < B2[2]) 
			{return true;}
		if ( (InBox( GetIntersection( L1[0]-B1[0], L2[0]-B1[0], L1, L2), B1, B2, 1 ))
		  || (InBox( GetIntersection( L1[1]-B1[1], L2[1]-B1[1], L1, L2), B1, B2, 2 )) 
		  || (InBox( GetIntersection( L1[2]-B1[2], L2[2]-B1[2], L1, L2), B1, B2, 3 )) 
		  || (InBox( GetIntersection( L1[0]-B2[0], L2[0]-B2[0], L1, L2), B1, B2, 1 )) 
		  || (InBox( GetIntersection( L1[1]-B2[1], L2[1]-B2[1], L1, L2), B1, B2, 2 )) 
		  || (InBox( GetIntersection( L1[2]-B2[2], L2[2]-B2[2], L1, L2), B1, B2, 3 )))
		  {
			return true;
		  }

		return false;
	}
	// -------------------------------------------------------------------------
	// Ray-Triangle Intersection Test
	function intersect_triangle(orig, dir, vert0, vert1, vert2) {
		var EPSILON = 0.000001;
		var edge1, edge2, tvec, pvec, qvec;
		var det, inv_det;
		var t, u, v;

		/* find vectors for two edges sharing vert0 */
		edge1 = subtractVectors(vert1, vert0);
		edge2 = subtractVectors(vert2, vert0);

		/* begin calculating determinant - also used to calculate U parameter */
		pvec = vectorCrossProduct(dir, edge2);

		/* if determinant is near zero, ray lies in plane of triangle */
		det = vectorDotProduct(edge1, pvec);

		if (TEST_CULL) {
			if (det < EPSILON)
			  return false;

			/* calculate distance from vert0 to ray origin */
			tvec = subtractVectors(orig, vert0);

			/* calculate U parameter and test bounds */
			u = vectorDotProduct(tvec, pvec);
			if (u < 0.0 || u > det)
			  return false;

			/* prepare to test V parameter */
			qvec = vectorCrossProduct(tvec, edge1);

			/* calculate V parameter and test bounds */
			v = vectorDotProduct(dir, qvec);
			if (v < 0.0 || u + v > det)
			  return false;

			/* calculate t, scale parameters, ray intersects triangle */
			t = vectorDotProduct(edge2, qvec);
			inv_det = 1.0 / det;
			t *= inv_det;
			u *= inv_det;
			v *= inv_det;
		}
		else {
			if (det > -EPSILON && det < EPSILON)
			 return false;
			inv_det = 1.0 / det;

			/* calculate distance from vert0 to ray origin */
			tvec = subtractVectors(orig, vert0);

			/* calculate U parameter and test bounds */
			u = vectorDotProduct(tvec, pvec) * inv_det;
			if (u < 0.0 || u > 1.0)
			 return false;

			/* prepare to test V parameter */
			qvec = vectorCrossProduct(tvec, edge1);

			/* calculate V parameter and test bounds */
			v = vectorDotProduct(dir, qvec) * inv_det;
			if (v < 0.0 || u + v > 1.0)
			 return false;

			/* calculate t, ray intersects triangle */
			t = vectorDotProduct(edge2, qvec) * inv_det;
		}
		return true;
	}

	function intersect_triangle2(orig, dir, vert0, vert1, vert2) {
		var edge1 = subtractVectors(vert1, vert0);
		var edge2 = subtractVectors(vert2, vert0);

		/* begin calculating determinant - also used to calculate U parameter */
		var pvec = vectorCrossProduct(dir, edge2);

		/* if determinant is near zero, ray lies in plane of triangle */
		var det = Math.abs(vectorDotProduct(edge1, pvec)); // = Vec3U::TripleProduct(dir,edge1,edge2);

		/* calculate distance from vert0 to ray origin */
		var tvec = subtractVectors(orig, vert0);

		/* calculate U parameter and test bounds */
		var u = vectorDotProduct(tvec, pvec); // = Vec3U::TripleProduct(dir,edge2,tvec);
		if ( u < 0 || u > det )
			return false;

		/* calculate V parameter and test bounds */
		var v = vectorDotProduct(dir, vectorCrossProduct(tvec,edge1)); // = dir * (tvec ^ edge1)
		if ( v < 0 || u + v > det )
			return false;

		return true;
	}
}